# -*- coding: utf-8 -*-
"""
===========================================================================
Nanotube structure generators (:mod:`sknano.nanogen._nanotube_generators`)
===========================================================================

.. currentmodule:: sknano.nanogen._nanotube_generators

.. todo::

   Add methods to perform fractional translation and cartesian translation
   before structure generation.

.. todo::

   Handle different units in output coordinates.

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import copy
#import itertools
#import sys

import numpy as np

from ..chemistry import Atom, Atoms
from ..tools import plural_word_check
from ..tools.refdata import CCbond

from ._nanotubes import Nanotube
from ._structure_generator import StructureGenerator

__all__ = ['NanotubeGenerator', 'UnrolledNanotubeGenerator', 'MWNTGenerator']


class NanotubeGenerator(Nanotube, StructureGenerator):
    u"""Class for generating nanotube structures.

    Parameters
    ----------
    n, m : int
        Chiral indices defining the nanotube chiral vector
        :math:`\\mathbf{C}_{h} = n\\mathbf{a}_{1} + m\\mathbf{a}_{2} = (n, m)`.
    nz : int, optional
        Number of repeat unit cells in the :math:`z` direction, along
        the *length* of the nanotube.
    element1, element2 : {str, int}, optional
        Element symbol or atomic number of basis
        :class:`~sknano.chemistry.Atoms` 1 and 2
    bond : float, optional
        :math:`\\mathrm{a}_{\\mathrm{CC}} =` distance between
        nearest neighbor atoms. Must be in units of **Angstroms**.
    Lz : float, optional
        Length of nanotube in units of **nanometers**.
        Overrides the `nz` value.

        .. versionadded:: 0.2.5

    tube_length : float, optional
        Length of nanotube in units of **nanometers**.
        Overrides the `nz` value.

        .. deprecated:: 0.2.5
           Use `Lz` instead

    fix_Lz : bool, optional
        Generate the nanotube with length as close to the specified
        :math:`L_z` as possible. If `True`, then
        non integer :math:`n_z` cells are permitted.

        .. versionadded:: 0.2.6

    autogen : bool, optional
        if `True`, automatically call
        :meth:`~NanotubeGenerator.generate_unit_cell`,
        followed by :meth:`~NanotubeGenerator.generate_structure_data`.
    with_units : bool, optional
        Attach `units` to physical quantities
    units : None, optional
        System of units to use.
    verbose : bool, optional
        if `True`, show verbose output

    Examples
    --------
    First, load the :class:`~sknano.nanogen.NanotubeGenerator` class.

    >>> from sknano.nanogen import NanotubeGenerator

    Now let's generate a :math:`\\mathbf{C}_{\\mathrm{h}} = (10, 5)`
    SWCNT unit cell.

    >>> nt = NanotubeGenerator(n=10, m=5)
    >>> nt.save_data(fname='10,5_unit_cell.xyz')

    The rendered structure looks like (orhographic view):

    .. image:: /images/10,5_unit_cell_orthographic_view.png

    and the perspective view:

    .. image:: /images/10,5_unit_cell_perspective_view.png

    """

    def __init__(self, n=int, m=int, nx=1, ny=1, nz=1,
                 element1='C', element2='C', bond=CCbond, tube_length=None,
                 Lx=None, Ly=None, Lz=None, fix_Lz=False,
                 with_units=False, units=None,
                 autogen=True, verbose=False):

        if tube_length is not None and Lz is None:
            Lz = tube_length

        super(NanotubeGenerator, self).__init__(
            n=n, m=m, nx=nx, ny=ny, nz=nz,
            element1=element1, element2=element2,
            bond=bond, Lx=Lx, Ly=Ly, Lz=Lz, fix_Lz=fix_Lz,
            with_units=False, units=units, verbose=verbose)

        if autogen:
            self.generate_unit_cell()
            self.generate_structure_data()

    def generate_unit_cell(self):
        """Generate the nanotube unit cell."""
        eps = 0.01
        n = self._n
        m = self._m
        bond = self._bond
        M = self._M
        T = self._T
        N = self._N
        rt = self._rt
        e1 = self._element1
        e2 = self._element2
        verbose = self._verbose

        aCh = Nanotube.compute_chiral_angle(n=n, m=m, rad2deg=False)

        tau = M * T / N
        dtau = bond * np.sin(np.pi / 6 - aCh)

        psi = 2 * np.pi / N
        dpsi = bond * np.cos(np.pi / 6 - aCh) / rt

        if verbose:
            print('dpsi: {}'.format(dpsi))
            print('dtau: {}\n'.format(dtau))

        self._unit_cell = Atoms()

        for i in xrange(1, N + 1):
            x1 = rt * np.cos(i * psi)
            y1 = rt * np.sin(i * psi)
            z1 = i * tau

            while z1 > T - eps:
                z1 -= T

            atom1 = Atom(e1, x=x1, y=y1, z=z1)
            atom1.rezero_coords()

            if verbose:
                print('Basis Atom 1:\n{}'.format(atom1))

            self._unit_cell.append(atom1)

            x2 = rt * np.cos(i * psi + dpsi)
            y2 = rt * np.sin(i * psi + dpsi)
            z2 = i * tau - dtau
            while z2 > T - eps:
                z2 -= T

            atom2 = Atom(e2, x=x2, y=y2, z=z2)
            atom2.rezero_coords()

            if verbose:
                print('Basis Atom 2:\n{}'.format(atom2))

            self._unit_cell.append(atom2)

    def generate_structure_data(self):
        """Generate structure data."""
        self._structure_atoms = Atoms()
        for nz in xrange(int(np.ceil(self._nz))):
            dr = np.array([0.0, 0.0, nz * self.T])
            for uc_atom in self._unit_cell:
                nt_atom = Atom(uc_atom.symbol)
                nt_atom.r = uc_atom.r + dr
                self._structure_atoms.append(nt_atom)

    def save_data(self, fname=None, outpath=None, structure_format=None,
                  rotation_angle=None, rot_axis=None, deg2rad=True,
                  center_CM=True):
        """Save structure data.

        See :meth:`~sknano.nanogen.StructureGenerator.save_data` method
        for documentation.

        """
        if fname is None:
            chirality = '{}{}r'.format('{}'.format(self._n).zfill(2),
                                       '{}'.format(self._m).zfill(2))
            if self._assume_integer_unit_cells:
                nz = ''.join(('{}'.format(self._nz),
                              plural_word_check('cell', self._nz)))
            else:
                nz = ''.join(('{:.2f}'.format(self._nz),
                              plural_word_check('cell', self._nz)))
            fname_wordlist = (chirality, nz)
            fname = '_'.join(fname_wordlist)

        if center_CM:
            self._structure_atoms.center_CM()

        if self._L0 is not None and self._fix_Lz:
            self._structure_atoms.clip_bounds(
                min_limit=-(10 * self._L0 + 0.2) / 2, r_indices=[2])
            if center_CM:
                self._structure_atoms.center_CM()
            self._structure_atoms.clip_bounds(
                max_limit=(10 * self._L0 + 0.2) / 2, r_indices=[2])
            if center_CM:
                self._structure_atoms.center_CM()

        super(NanotubeGenerator, self).save_data(
            fname=fname, outpath=outpath, structure_format=structure_format,
            rotation_angle=rotation_angle, rot_axis=rot_axis,
            deg2rad=deg2rad, center_CM=False)


class CappedNanotubeGenerator(Nanotube, StructureGenerator):
    u"""Class for generating capped nanotube structures."""
    pass


class UnrolledNanotubeGenerator(Nanotube, StructureGenerator):
    u"""Class for generating unrolled nanotube structures.

    .. versionadded:: 0.2.23

    Parameters
    ----------
    n, m : int
        Chiral indices defining the nanotube chiral vector
        :math:`\\mathbf{C}_{h} = n\\mathbf{a}_{1} + m\\mathbf{a}_{2} = (n, m)`.
    nx, ny, nz : int, optional
        Number of repeat unit cells in the :math:`x, y, z` dimensions
    element1, element2 : {str, int}, optional
        Element symbol or atomic number of basis
        :class:`~sknano.chemistry.Atoms` 1 and 2
    bond : float, optional
        :math:`\\mathrm{a}_{\\mathrm{CC}} =` distance between
        nearest neighbor atoms. Must be in units of **Angstroms**.
    Lx, Ly, Lz : float, optional
        Length of bundle in :math:`x, y, z` dimensions in **nanometers**.
        Overrides the :math:`n_x, n_y, n_z` cell values.
    fix_Lz : bool, optional
        Generate the nanotube with length as close to the specified
        :math:`L_z` as possible. If `True`, then
        non integer :math:`n_z` cells are permitted.
    autogen : bool, optional
        if `True`, automatically call
        :meth:`~NanotubeGenerator.generate_unit_cell`,
        followed by :meth:`~NanotubeGenerator.generate_structure_data`.
    with_units : bool, optional
        Attach `units` to physical quantities
    units : None, optional
        System of units to use.
    verbose : bool, optional
        if `True`, show verbose output

    Notes
    -----
    The `UnrolledNanotubeGenerator` class generates graphene using the
    nanotube unit cell defined by the chiral vector
    :math:`\\mathbf{C}_{h} = n\\mathbf{a}_{1} + m\\mathbf{a}_{2} = (n, m)`.
    If you want to generate graphene with an armchair or zigzag edge using
    `length` and `width` parameters, see the
    :class:`~sknano.nanogen.GrapheneGenerator` class.

    .. seealso:: :class:`~sknano.nanogen.GrapheneGenerator`


    Examples
    --------
    First, load the :class:`~sknano.nanogen.UnrolledNanotubeGenerator`
    class.

    >>> from sknano.nanogen import UnrolledNanotubeGenerator

    Now let's generate an unrolled :math:`\\mathbf{C}_{\\mathrm{h}} = (10, 5)`
    SWCNT unit cell.

    >>> flatswcnt = UnrolledNanotubeGenerator(n=10, m=5)
    >>> flatswcnt.save_data()

    The rendered structure looks like:

    """

    def __init__(self, n=int, m=int, nx=1, ny=1, nz=1,
                 element1='C', element2='C', bond=CCbond,
                 Lx=None, Ly=None, Lz=None, fix_Lz=False,
                 nlayers=None, layer_spacing=None, stacking_order=None,
                 with_units=False, units=None, autogen=True, verbose=False):

        super(UnrolledNanotubeGenerator, self).__init__(
            n=n, m=m, nx=nx, ny=ny, nz=nz,
            element1=element1, element2=element2,
            bond=bond, Lx=Lx, Ly=Ly, Lz=Lz, fix_Lz=fix_Lz,
            with_units=False, units=units, verbose=verbose)

        if autogen:
            self.generate_unit_cell()
            self.generate_structure_data()

    def generate_unit_cell(self):
        """Generate the nanotube unit cell."""
        eps = 0.01
        n = self._n
        m = self._m
        bond = self._bond
        M = self._M
        T = self._T
        N = self._N
        rt = self._rt
        e1 = self._element1
        e2 = self._element2
        verbose = self._verbose

        aCh = Nanotube.compute_chiral_angle(n=n, m=m, rad2deg=False)

        tau = M * T / N
        dtau = bond * np.sin(np.pi / 6 - aCh)

        psi = 2 * np.pi / N
        dpsi = bond * np.cos(np.pi / 6 - aCh) / rt

        if verbose:
            print('dpsi: {}'.format(dpsi))
            print('dtau: {}\n'.format(dtau))

        self._unit_cell = Atoms()

        for i in xrange(N):
            x1 = rt * i * psi
            z1 = i * tau

            while z1 > T - eps:
                z1 -= T

            atom1 = Atom(e1, x=x1, z=z1)
            atom1.rezero_coords()

            if verbose:
                print('Basis Atom 1:\n{}'.format(atom1))

            self._unit_cell.append(atom1)

            x2 = rt * (i * psi + dpsi)
            z2 = i * tau - dtau
            while z2 > T - eps:
                z2 -= T

            atom2 = Atom(e2, x=x2, z=z2)
            atom2.rezero_coords()

            if verbose:
                print('Basis Atom 2:\n{}'.format(atom2))

            self._unit_cell.append(atom2)

    def generate_structure_data(self):
        """Generate structure data."""
        self._structure_atoms = Atoms()
        for nx in xrange(self.nx):
            for nz in xrange(int(np.ceil(self.nz))):
                dr = np.array([nx * self.Ch, 0.0, nz * self.T])
                for uc_atom in self._unit_cell:
                    nt_atom = Atom(uc_atom.symbol)
                    nt_atom.r = uc_atom.r + dr
                    self._structure_atoms.append(nt_atom)

    def save_data(self, fname=None, outpath=None, structure_format=None,
                  rotation_angle=None, rot_axis=None, deg2rad=True,
                  center_CM=True):
        """Save structure data.

        See :meth:`~sknano.nanogen.StructureGenerator.save_data` method
        for documentation.

        """
        if fname is None:
            chirality = '{}{}f'.format('{}'.format(self._n).zfill(2),
                                       '{}'.format(self._m).zfill(2))
            nx = self.nx
            ny = self.ny
            fname_wordlist = None
            if nx != 1 or ny != 1:
                nx = ''.join(('{}'.format(self.nx),
                              plural_word_check('cell', self.nx)))
                ny = ''.join(('{}'.format(self.ny),
                              plural_word_check('cell', self.ny)))

                if self._assume_integer_unit_cells:
                    nz = ''.join(('{}'.format(self.nz),
                                  plural_word_check('cell', self.nz)))
                else:
                    nz = ''.join(('{:.2f}'.format(self.nz),
                                  plural_word_check('cell', self.nz)))

                cells = 'x'.join((nx, ny, nz))
                fname_wordlist = (chirality, cells)
            else:
                if self._assume_integer_unit_cells:
                    nz = ''.join(('{}'.format(self.nz),
                                  plural_word_check('cell', self.nz)))
                else:
                    nz = ''.join(('{:.2f}'.format(self.nz),
                                  plural_word_check('cell', self.nz)))

                fname_wordlist = (chirality, nz)

            fname = '_'.join(fname_wordlist)

        if center_CM:
            self._structure_atoms.center_CM()

        if self._L0 is not None and self._fix_Lz:
            self._structure_atoms.clip_bounds(
                min_limit=-(10 * self._L0 + 0.2) / 2, r_indices=[2])
            if center_CM:
                self._structure_atoms.center_CM()
            self._structure_atoms.clip_bounds(
                max_limit=(10 * self._L0 + 0.2) / 2, r_indices=[2])
            if center_CM:
                self._structure_atoms.center_CM()

        super(UnrolledNanotubeGenerator, self).save_data(
            fname=fname, outpath=outpath, structure_format=structure_format,
            rotation_angle=rotation_angle, rot_axis=rot_axis,
            deg2rad=deg2rad, center_CM=False)


class MWNTGenerator(NanotubeGenerator):
    u"""Class for generating single, multi-walled nanotubes.

    .. versionchanged:: 0.2.20

       `MWNTGenerator` no longer generates MWNT *bundles*, only *single*
       MWNTs. To generate bundled MWNT structure data, use the
       `MWNTBundleGenerator` class.

    .. versionadded:: 0.2.8

    Parameters
    ----------
    n, m : int
        Chiral indices defining the nanotube chiral vector
        :math:`\\mathbf{C}_{h} = n\\mathbf{a}_{1} + m\\mathbf{a}_{2} = (n, m)`.
    nx, ny, nz : int, optional
        Number of repeat unit cells in the :math:`x, y, z` dimensions.
    element1, element2 : {str, int}, optional
        Element symbol or atomic number of basis
        :class:`~sknano.chemistry.Atoms` 1 and 2
    bond : float, optional
        :math:`\\mathrm{a}_{\\mathrm{CC}} =` distance between
        nearest neighbor atoms. Must be in units of **Angstroms**.
    Lx, Ly, Lz : float, optional
        length of bundle in :math:`x, y, z` dimensions in **nanometers**.
        Overrides the :math:`n_x, n_y, n_z` cell values.
    fix_Lz : bool, optional
        Generate the nanotube with length as close to the specified
        :math:`L_z` as possible. If `True`, then
        non integer :math:`n_z` cells are permitted.
    add_outer_shells : bool, optional
        Build the MWNT by adding outer shells

        .. versionadded:: 0.2.23

    add_inner_shells : bool, optional
        Build the MWNT by adding inner shells

        .. versionadded:: 0.2.23

    max_shells : int, optional
        Maximum number of shells per MWNT.
    max_shell_diameter : float, optional
        Maximum shell diameter, in units of **Angstroms**.
    min_shells : int, optional
        Minimum number of shells per MWNT.
    min_shell_diameter : float, optional
        Minimum shell diameter, in units of **Angstroms**.
    new_shell_type : {None, 'AC', 'ZZ', 'achiral', 'chiral'}, optional
        If `None`, the chiralities of the new shells are constrained only
        by their diameter and will be chosen randomly if more than one
        candidate chirality exists. If not `None`, then the
        `new_shell_type` will be added as a constraint.
    shell_spacing : float, optional
        Shell spacing in units of **Angstroms**. Default
        value is the van der Waals interaction distance of 3.4 Angstroms.
    autogen : bool, optional
        if `True`, automatically call
        :meth:`~MWNTGenerator.generate_unit_cell`,
        followed by :meth:`~MWNTGenerator.generate_structure_data`.
    verbose : bool, optional
        if `True`, show verbose output

    Examples
    --------

    >>> from sknano.nanogen import MWNTGenerator
    >>> mwnt = MWNTGenerator(n=40, m=40, max_shells=5, Lz=1.0, fix_Lz=True)
    >>> mwnt.save_data()

    .. image:: /images/5shell_mwnt_4040_outer_Ch_1cellx1cellx4.06cells-01.png

    """
    def __init__(self, n=int, m=int, nx=1, ny=1, nz=1,
                 element1='C', element2='C', bond=CCbond,
                 Lx=None, Ly=None, Lz=None, fix_Lz=False,
                 add_inner_shells=True, add_outer_shells=False,
                 max_shells=None, max_shell_diameter=np.inf,
                 min_shells=None, min_shell_diameter=0.0,
                 new_shell_type=None, shell_spacing=3.4,
                 with_units=False, units=None, autogen=True, verbose=False):

        super(MWNTGenerator, self).__init__(
            n=n, m=m, nx=nx, ny=ny, nz=nz,
            element1=element1, element2=element2, bond=bond,
            Lx=Lx, Ly=Ly, Lz=Lz, fix_Lz=fix_Lz,
            with_units=with_units, units=units,
            autogen=False, verbose=verbose)

        self._add_inner_shells = add_inner_shells
        self._add_outer_shells = add_outer_shells
        self._starting_shell_position = 'outer'

        self._max_shells = max_shells
        if max_shells is None:
            self._max_shells = 10
        self._max_shell_diameter = max_shell_diameter

        self._min_shells = min_shells
        if min_shells is None:
            self._min_shells = 2
        self._min_shell_diameter = min_shell_diameter

        self._new_shell_type = new_shell_type
        self._shell_spacing = shell_spacing

        self._Nshells_per_tube = 1
        self._Natoms_per_tube = 0

        if autogen:
            super(MWNTGenerator, self).generate_unit_cell()
            self.generate_structure_data()

    def _generate_unit_cell(self, n=int, m=int):
        """Generate the unit cell of a MWNT shell"""
        eps = 0.01
        bond = self._bond
        e1 = self._element1
        e2 = self._element2

        N = Nanotube.compute_N(n=n, m=m)
        aCh = Nanotube.compute_chiral_angle(n=n, m=m, rad2deg=False)
        rt = Nanotube.compute_rt(n=n, m=m, bond=bond, with_units=False)
        T = Nanotube.compute_T(n=n, m=m, bond=bond, with_units=False)

        tau = Nanotube.compute_tau(n=n, m=m, bond=bond, with_units=False)
        dtau = bond * np.sin(np.pi / 6 - aCh)

        psi = Nanotube.compute_psi(n=n, m=m)
        dpsi = bond * np.cos(np.pi / 6 - aCh) / rt

        unit_cell = Atoms()

        for i in xrange(1, N + 1):
            x1 = rt * np.cos(i * psi)
            y1 = rt * np.sin(i * psi)
            z1 = i * tau

            while z1 > T - eps:
                z1 -= T

            atom1 = Atom(e1, x=x1, y=y1, z=z1)
            atom1.rezero_coords()

            unit_cell.append(atom1)

            x2 = rt * np.cos(i * psi + dpsi)
            y2 = rt * np.sin(i * psi + dpsi)
            z2 = i * tau - dtau
            while z2 > T - eps:
                z2 -= T

            atom2 = Atom(e2, x=x2, y=y2, z=z2)
            atom2.rezero_coords()

            unit_cell.append(atom2)

        return unit_cell

    def generate_structure_data(self):
        """Generate structure data.

        .. todo::

           Load the diameter and chirality data from file instead of
           generating it every time.

        """
        Ch = []
        dt = []
        for n in xrange(0, 201):
            for m in xrange(0, 201):
                if (n <= 2 and m <= 2):
                    continue
                else:
                    Ch.append((n, m))
                    dt.append(Nanotube.compute_dt(n=n, m=m, bond=self._bond))
        Ch = np.asarray(Ch)
        dt = np.asarray(dt)

        super(MWNTGenerator, self).generate_structure_data()

        swnt0 = copy.deepcopy(self._structure_atoms)
        self._structure_atoms = Atoms(atoms=swnt0, deepcopy=True)
        self._structure_atoms.center_CM()

        self._max_shell_diameter = min(self._max_shell_diameter, dt.max())
        self._min_shell_diameter = max(self._min_shell_diameter, dt.min())

        Lzmin = self.Lz

        delta_dt = -2 * self._shell_spacing
        max_dt_diff = 0.05

        if self._add_outer_shells:
            delta_dt = -delta_dt
            self._starting_shell_position = 'inner'

        next_dt = self.dt + delta_dt
        while self._Nshells_per_tube < self._max_shells and \
                next_dt <= self._max_shell_diameter and \
                next_dt >= self._min_shell_diameter:

            # get chiral indices for next_dt
            next_Ch_candidates = []
            while len(next_Ch_candidates) == 0 and \
                    next_dt <= self._max_shell_diameter and \
                    next_dt >= self._min_shell_diameter:

                if self._new_shell_type in ('AC', 'armchair'):
                    next_Ch_candidates = \
                        Ch[np.where(
                            np.logical_and(np.abs(dt - next_dt) <= max_dt_diff,
                                           Ch[:,0] == Ch[:,1]))]
                elif self._new_shell_type in ('ZZ', 'zigzag'):
                    next_Ch_candidates = \
                        Ch[np.where(
                            np.logical_and(np.abs(dt - next_dt) <= max_dt_diff,
                                           np.logical_or(Ch[:,0] == 0,
                                                         Ch[:,1] == 0)))]
                elif self._new_shell_type == 'achiral':
                    next_Ch_candidates = \
                        Ch[np.where(
                            np.logical_and(np.abs(dt - next_dt) <= max_dt_diff,
                                           np.logical_or(
                                               Ch[:,0] == Ch[:,1],
                                               np.logical_or(
                                                   Ch[:,0] == 0,
                                                   Ch[:,1] == 0))))]
                elif self._new_shell_type == 'chiral':
                    next_Ch_candidates = \
                        Ch[np.where(
                            np.logical_and(np.abs(dt - next_dt) <= max_dt_diff,
                                           np.logical_and(
                                               Ch[:,0] != Ch[:,1],
                                               np.logical_and(
                                                   Ch[:,0] != 0,
                                                   Ch[:,1] != 1))))]
                else:
                    next_Ch_candidates = \
                        Ch[np.where(np.abs(dt - next_dt) <= max_dt_diff)]

                if self._add_outer_shells:
                    next_dt += max_dt_diff
                else:
                    next_dt -= max_dt_diff

            if len(next_Ch_candidates) > 0:
                n, m = next_Ch_candidates[
                    np.random.choice(np.arange(len(next_Ch_candidates)))]
                T = Nanotube.compute_T(n=n, m=m, bond=self._bond)
                Lz = Nanotube.compute_Lz(
                    n=n, m=m, bond=self._bond, nz=self._nz)
                Lzmin = min(Lzmin, Lz)

                # generate unit cell for new shell chiral indices
                shell_unit_cell = self._generate_unit_cell(n=n, m=m)

                if self._verbose:
                    print('new shell:\n'
                          'n, m = {}, {}\n'.format(n, m) +
                          'dt: {:.4f}\n'.format(next_dt) +
                          'shell_unit_cell.Natoms: {}\n'.format(
                              shell_unit_cell.Natoms))

                shell = Atoms()
                for nz in xrange(int(np.ceil(self._nz))):
                    dr = np.array([0.0, 0.0, nz * T])
                    for uc_atom in shell_unit_cell:
                        atom = Atom(uc_atom.symbol)
                        atom.r = uc_atom.r + dr
                        shell.append(atom)
                shell.center_CM()
                self._structure_atoms.extend(shell.atoms)
                self._Nshells_per_tube += 1
                next_dt += delta_dt
            else:
                break

        if self._L0 is not None and self._fix_Lz:
            self._structure_atoms.clip_bounds(
                abs_limit=(10 * self._L0 + 1) / 2, r_indices=[2])
        else:
            self._structure_atoms.clip_bounds(
                abs_limit=(10 * Lzmin + 1) / 2, r_indices=[2])

        self._Natoms_per_tube = self._structure_atoms.Natoms

        if self._verbose:
            print('Nshells_per_tube: {}'.format(self._Nshells_per_tube))
            print('Natoms_per_tube: {}'.format(self._Natoms_per_tube))

    def save_data(self, fname=None, outpath=None, structure_format=None,
                  rotation_angle=None, rot_axis=None, deg2rad=True,
                  center_CM=True):
        """Save structure data.

        See :meth:`~sknano.nanogen.StructureGenerator.save_data` method
        for documentation.

        """
        if fname is None:
            Nshells = '{}shell_mwnt'.format(self._Nshells_per_tube)
            chirality = '{}{}_{}_Ch'.format('{}'.format(self._n).zfill(2),
                                            '{}'.format(self._m).zfill(2),
                                            self._starting_shell_position)
            fname_wordlist = None
            if self._assume_integer_unit_cells:
                nz = ''.join(('{}'.format(self._nz),
                              plural_word_check('cell', self._nz)))
            else:
                nz = ''.join(('{:.2f}'.format(self._nz),
                              plural_word_check('cell', self._nz)))
            fname_wordlist = (Nshells, chirality, nz)
            fname = '_'.join(fname_wordlist)

        super(MWNTGenerator, self).save_data(
            fname=fname, outpath=outpath, structure_format=structure_format,
            rotation_angle=rotation_angle, rot_axis=rot_axis,
            deg2rad=deg2rad, center_CM=center_CM)



# -*- coding: utf-8 -*-
"""
==================================================================
Abstract data structures for math (:mod:`sknano.tools._coremath`)
==================================================================

.. currentmodule:: sknano.tools._coremath

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

try:
    from pint import UnitRegistry
    ureg = UnitRegistry()
    Qty = ureg.Quantity
except ImportError:
    Qty = None

import numpy as np

from ._corefuncs import check_type

__all__ = ['Point', 'Point2D', 'Point3D',
           'Vector', 'Vector2D', 'Vector3D',
           'Quaternion']


class Point(object):
    """Create a point in :math:`R^3`

    Parameters
    ----------
    x, y, z : float, optional
        :math:`x, y, z` coordinates of point in :math:`R^3` space.
    units : {None, str}, optional
        Units of coordinates.

    """
    def __init__(self, x=None, y=None, z=None, with_units=False,
                 units=None, dtype=None):
        self._p = np.zeros(3, dtype=float)
        self._units = units
        for i, pi in enumerate((x, y, z)):
            if pi is not None:
                self._p[i] = pi

        if with_units and (Qty is None or units is None):
            with_units = False
        self._with_units = with_units
        self._units = units

        if with_units:
            self._p = Qty(self._p, units)

    def __str__(self):
        if self._with_units:
            return str(self._p)
        else:
            return '({}, {}, {})'.format(self.x, self.y, self.z)

    def __repr__(self):
        if self._with_units:
            return repr(self._p)
        else:
            return '({}, {}, {})'.format(self.x, self.y, self.z)

    @property
    def x(self):
        """:math:`x`-coordinate of `Point`.

        Returns
        -------
        float
            :math:`x`-coordinate of `Point`.

        """
        return self._p[0]

    @x.setter
    def x(self, value=float):
        """Set :math:`x`-coordinate of `Point`.

        Parameters
        ----------
        value : float
            :math:`x`-coordinate of `Point`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(int, float))
            try:
                self._p[0] = value
            except ValueError:
                self._p[0] = Qty(value, self._units)
        except TypeError as e:
            print(e)

    @property
    def y(self):
        """:math:`y`-coordinate of `Point`.

        Returns
        -------
        float
            :math:`y`-coordinate of `Point`.

        """
        return self._p[1]

    @y.setter
    def y(self, value=float):
        """Set :math:`y`-coordinate of `Point`.

        Parameters
        ----------
        value : float
            :math:`y`-coordinate of `Point`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(int, float))
            try:
                self._p[1] = value
            except ValueError:
                self._p[1] = Qty(value, self._units)
        except TypeError as e:
            print(e)

    @property
    def z(self):
        """:math:`z`-coordinate of `Point`.

        Returns
        -------
        float
            :math:`z`-coordinate of `Point`.

        """
        return self._p[2]

    @z.setter
    def z(self, value=float):
        """Set :math:`z`-coordinate of `Point`.

        Parameters
        ----------
        value : float
            :math:`z`-coordinate of `Point`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(int, float))
            try:
                self._p[2] = value
            except ValueError:
                self._p[2] = Qty(value, self._units)
        except TypeError as e:
            print(e)

    @property
    def coords(self):
        """:math:`x, y, z` coordinates of `Point`.

        Returns
        -------
        :py:class:`~numpy:numpy.ndarray`
            3-element :py:class:`~numpy:numpy.ndarray` of [:math:`x, y, z`]
            coordinates of `Point`.

        """
        return self._p

    @coords.setter
    def coords(self, value=np.ndarray):
        """Set :math:`x, y, z` coordinates of `Point`

        Parameters
        ----------
        value : :py:class:`~numpy:numpy.ndarray`
            3-element :py:class:`~numpy:numpy.ndarray` of [:math:`x, y, z`]
            coordinates of `Point`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(np.ndarray,))
            for i, pi in enumerate(value):
                try:
                    check_type(pi, allowed_types=(int, float))
                    try:
                        self._p[i] = pi
                    except ValueError:
                        self._p[i] = Qty(pi, self._units)
                except TypeError as e:
                    print(e)
        except TypeError as e:
            print(e)

    def fix_minus_zero_coords(self, epsilon=1.0e-10):
        """Set `Point` coordinates that are small, negative numbers to zero.

        Set `Point` coordinates that are negative and have absolute value
        less than `epsilon` to zero.

        Parameters
        ----------
        epsilon : float, optional
            Smallest allowed absolute value of any :math:`x,y,z` coordinate.

        """
        p = self._p.tolist()
        for i, pi in enumerate(p[:]):
            if pi < 0 and abs(pi) < epsilon:
                try:
                    self._p[i] = 0.0
                except ValueError:
                    self._p[i] = Qty(0.0, self._units)

    def rezero_coords(self, epsilon=1.0e-10):
        """Re-zero `Point` coordinates near zero.

        Set `Point` coordinates with absolute value less than `epsilon` to
        zero.

        Parameters
        ----------
        epsilon : float, optional
            Smallest allowed absolute value of any :math:`x,y,z` coordinate.

        """
        p = self._p.tolist()
        for i, pi in enumerate(p[:]):
            if abs(pi) < epsilon:
                try:
                    self._p[i] = 0.0
                except ValueError:
                    self._p[i] = Qty(0.0, self._units)


Point3D = Point


class Point2D(object):
    """Create a point in :math:`R^2`

    Parameters
    ----------
    x, y : float, optional
        :math:`x, y` coordinates of point in :math:`R^2` space.
    units : {None, str}, optional
        Units of coordinates.

    """
    def __init__(self, x=None, y=None, with_units=False, units=None,
                 dtype=None):
        self._p = np.zeros(2, dtype=float)
        self._units = units
        for i, pi in enumerate((x, y)):
            if pi is not None:
                self._p[i] = pi

        if with_units and (Qty is None or units is None):
            with_units = False
        self._with_units = with_units
        self._units = units

        if with_units:
            self._p = Qty(self._p, units)

    def __str__(self):
        if self._with_units:
            return str(self._p)
        else:
            return '({}, {})'.format(self.x, self.y)

    def __repr__(self):
        if self._with_units:
            return repr(self._p)
        else:
            return '({}, {})'.format(self.x, self.y)

    @property
    def x(self):
        """:math:`x`-coordinate of `Point2D`.

        Returns
        -------
        float
            :math:`x`-coordinate of `Point2D`.

        """
        return self._p[0]

    @x.setter
    def x(self, value=float):
        """Set :math:`x`-coordinate of `Point2D`.

        Parameters
        ----------
        value : float
            :math:`x`-coordinate of `Point2D`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(int, float))
            try:
                self._p[0] = value
            except ValueError:
                self._p[0] = Qty(value, self._units)
        except TypeError as e:
            print(e)

    @property
    def y(self):
        """:math:`y`-coordinate of `Point2D`.

        Returns
        -------
        float
            :math:`y`-coordinate of `Point2D`.

        """
        return self._p[1]

    @y.setter
    def y(self, value=float):
        """Set :math:`y`-coordinate of `Point2D`.

        Parameters
        ----------
        value : float
            :math:`y`-coordinate of `Point2D`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(int, float))
            try:
                self._p[1] = value
            except ValueError:
                self._p[1] = Qty(value, self._units)
        except TypeError as e:
            print(e)

    @property
    def coords(self):
        """:math:`x, y` coordinates of `Point2D`.

        Returns
        -------
        :py:class:`~numpy:numpy.ndarray`
            2-element :py:class:`~numpy:numpy.ndarray` of [:math:`x, y`]
            coordinates of `Point2D`.

        """
        return self._p

    @coords.setter
    def coords(self, value=np.ndarray):
        """Set :math:`x, y` coordinates of `Point2D`

        Parameters
        ----------
        value : :py:class:`~numpy:numpy.ndarray`
            2-element :py:class:`~numpy:numpy.ndarray` of [:math:`x, y`]
            coordinates of `Point2D`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(np.ndarray,))
            for i, pi in enumerate(value):
                try:
                    check_type(pi, allowed_types=(int, float))
                    try:
                        self._p[i] = pi
                    except ValueError:
                        self._p[i] = Qty(pi, self._units)
                except TypeError as e:
                    print(e)
        except TypeError as e:
            print(e)

    def fix_minus_zero_coords(self, epsilon=1.0e-10):
        """Set `Point2D` coordinates that are small, negative numbers to zero.

        Set `Point2D` coordinates that are negative and have absolute value
        less than `epsilon` to zero.

        Parameters
        ----------
        epsilon : float, optional
            Smallest allowed absolute value of any :math:`x,y` coordinate.

        """
        p = self._p.tolist()
        for i, pi in enumerate(p[:]):
            if pi < 0 and abs(pi) < epsilon:
                try:
                    self._p[i] = 0.0
                except ValueError:
                    self._p[i] = Qty(0.0, self._units)

    def rezero_coords(self, epsilon=1.0e-10):
        """Re-zero `Point2D` coordinates near zero.

        Set `Point2D` coordinates with absolute value less than `epsilon` to
        zero.

        Parameters
        ----------
        epsilon : float, optional
            Smallest allowed absolute value of any :math:`x,y` coordinate.

        """
        p = self._p.tolist()
        for i, pi in enumerate(p[:]):
            if abs(pi) < epsilon:
                try:
                    self._p[i] = 0.0
                except ValueError:
                    self._p[i] = Qty(0.0, self._units)


class Vector(object):
    """Create a vector in :math:`R^3`

    Parameters
    ----------
    x, y, z : float, optional
        :math:`x, y, z` components of terminating point of vector in
        :math:`R^3` space relative to origin.
    x0, y0, z0 : float, optional
        :math:`x_0, y_0, z_0` components of starting point of vector in
        :math:`R^3` space relative to origin.
    p, p0 : `Point`, optional
        Terminating and starting `Point` of vector in :math:`R^3` space
        relative to origin. If `p` is not `None` it will always override the
        `Point` defined by the `x`, `y`, `z` parameters. Similarly, if `p0` is
        not `None`, it will always override the `Point` defined by the
        `x0`, `y0`, `z0` parameters.
    units : {None, str}, optional
        Units of vector.


    Notes
    -----
    .. todo::

       add new methods for coordinate transformations


    """
    def __init__(self, x=None, y=None, z=None, x0=None, y0=None, z0=None,
                 p=None, p0=None, with_units=False, units=None):

        self._p = None
        if p is None:
            self._p = Point(x=x, y=y, z=z)
        elif isinstance(p, (Point, Point3D)):
            if isinstance(p.coords, np.ndarray):
                self._p = p
            else:
                p_coords = p.coords.magnitude
                self._p = Point()
                self._p.coords = p_coords

        self._p0 = None
        if p0 is None:
            self._p0 = Point(x=x0, y=y0, z=z0)
        elif isinstance(p0, (Point, Point3D)):
            if isinstance(p0.coords, np.ndarray):
                self._p0 = p0
            else:
                p0_coords = p0.coords.magnitude
                self._p0 = Point()
                self._p0.coords = p0_coords

        self._v = np.zeros(3, dtype=float)
        for i, (pi, pi0) in enumerate(zip(self._p.coords, self._p0.coords)):
            self._v[i] = pi - pi0

        if with_units and (Qty is None or units is None):
            with_units = False
        self._with_units = with_units
        self._units = units

        if with_units:
            self._v = Qty(self._v, units)

    def __str__(self):
        if self._with_units:
            return str(self._v)
        else:
            return '({}, {}, {})'.format(self.x, self.y, self.z)

    def __repr__(self):
        if self._with_units:
            return repr(self._v)
        else:
            return '({}, {}, {})'.format(self.x, self.y, self.z)

    @property
    def x(self):
        """:math:`x`-coordinate of `Vector`.

        Returns
        -------
        float
            :math:`x`-coordinate of `Vector`.

        """
        return self._v[0]

    @x.setter
    def x(self, value=float):
        """Set :math:`x`-coordinate of `Vector`.

        Parameters
        ----------
        value : float
            :math:`x`-coordinate of `Vector`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(int, float))
            try:
                self._v[0] = value
            except ValueError:
                self._v[0] = Qty(value, self._units)
        except TypeError as e:
            print(e)

    @property
    def y(self):
        """:math:`y`-coordinate of `Vector`.

        Returns
        -------
        float
            :math:`y`-coordinate of `Vector`.

        """
        return self._v[1]

    @y.setter
    def y(self, value=float):
        """Set :math:`y`-coordinate of `Vector`.

        Parameters
        ----------
        value : float
            :math:`y`-coordinate of `Vector`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(int, float))
            try:
                self._v[1] = value
            except ValueError:
                self._v[1] = Qty(value, self._units)
        except TypeError as e:
            print(e)

    @property
    def z(self):
        """:math:`z`-coordinate of `Vector`.

        Returns
        -------
        float
            :math:`z`-coordinate of `Vector`.

        """
        return self._v[2]

    @z.setter
    def z(self, value=float):
        """Set :math:`z`-coordinate of `Vector`.

        Parameters
        ----------
        value : float
            :math:`z`-coordinate of `Vector`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(int, float))
            try:
                self._v[2] = value
            except ValueError:
                self._v[2] = Qty(value, self._units)
        except TypeError as e:
            print(e)

    @property
    def components(self):
        """:math:`x, y, z` components of `Vector`.

        Returns
        -------
        :py:class:`~numpy:numpy.ndarray`
            3-element :py:class:`~numpy:numpy.ndarray` of [:math:`x, y, z`]
            components of `Vector`.

        """
        return self._v

    @components.setter
    def components(self, value=np.ndarray):
        """Set :math:`x, y, z` components of `Vector`

        Parameters
        ----------
        value : :py:class:`~numpy:numpy.ndarray`
            3-element :py:class:`~numpy:numpy.ndarray` of [:math:`x, y, z`]
            components of `Vector`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(np.ndarray,))
            for i, vi in enumerate(value):
                try:
                    check_type(vi, allowed_types=(int, float))
                    try:
                        self._v[i] = vi
                    except ValueError:
                        self._v[i] = Qty(vi, self._units)
                except TypeError as e:
                    print(e)
        except TypeError as e:
            print(e)

    def fix_minus_zero_components(self, epsilon=1.0e-10):
        """Set `Vector` components that are small, negative numbers to zero.

        Set `Vector` components that are negative and have absolute value
        less than `epsilon` to zero.

        Parameters
        ----------
        epsilon : float, optional
            Smallest allowed absolute value of any :math:`x,y,z` component.

        """
        v = self._v.tolist()
        for i, vi in enumerate(v[:]):
            if vi < 0 and abs(vi) < epsilon:
                try:
                    self._v[i] = 0.0
                except ValueError:
                    self._v[i] = Qty(0.0, self._units)

    def rezero_components(self, epsilon=1.0e-10):
        """Re-zero `Vector` components near zero.

        Set `Vector` components with absolute value less than `epsilon` to
        zero.

        Parameters
        ----------
        epsilon : float, optional
            Smallest allowed absolute value of any :math:`x,y,z` component.

        """
        v = self._v.tolist()
        for i, vi in enumerate(v[:]):
            if abs(vi) < epsilon:
                try:
                    self._v[i] = 0.0
                except ValueError:
                    self._v[i] = Qty(0.0, self._units)


Vector3D = Vector


class Vector2D(object):
    """Create a vector in :math:`R^2`

    Parameters
    ----------
    x, y : float, optional
        :math:`x, y` components of terminating point of vector in
        :math:`R^2` space relative to origin.
    x0, y0 : float, optional
        :math:`x_0, y_0` components of starting point of vector in
        :math:`R^2` space relative to origin.
    p, p0 : `Point2D`, optional
        Terminating and starting `Point2D` of vector in :math:`R^2` space
        relative to origin. If `p` is not `None` it will always override the
        `Point2D` defined by the `x`, `y` parameters. Similarly, if `p0` is
        not `None`, it will always override the `Point2D` defined by the
        `x0`, `y0` parameters.
    units : {None, str}, optional
        Units of vector.

    """
    def __init__(self, x=None, y=None, x0=None, y0=None, p=None, p0=None,
                 with_units=False, units=None):

        self._p = None
        if p is None:
            self._p = Point2D(x=x, y=y)
        elif isinstance(p, Point2D):
            self._p = p

        self._p0 = None
        if p0 is None:
            self._p0 = Point2D(x=x0, y=y0)
        elif isinstance(p0, Point2D):
            self._p0 = p0

        self._v = np.zeros(2, dtype=float)
        for i, (pi, pi0) in enumerate(zip(self._p.coords, self._p0.coords)):
            self._v[i] = pi - pi0

        if with_units and (Qty is None or units is None):
            with_units = False
        self._with_units = with_units
        self._units = units

        if with_units:
            self._v = Qty(self._v, units)

    def __str__(self):
        if self._with_units:
            return str(self._v)
        else:
            return '({}, {})'.format(self.x, self.y)

    def __repr__(self):
        if self._with_units:
            return repr(self._v)
        else:
            return '({}, {})'.format(self.x, self.y)

    @property
    def x(self):
        """:math:`x`-coordinate of `Vector2D`.

        Returns
        -------
        float
            :math:`x`-coordinate of `Vector2D`.

        """
        return self._v[0]

    @x.setter
    def x(self, value=float):
        """Set :math:`x`-coordinate of `Vector2D`.

        Parameters
        ----------
        value : float
            :math:`x`-coordinate of `Vector2D`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(int, float))
            try:
                self._v[0] = value
            except ValueError:
                self._v[0] = Qty(value, self._units)
        except TypeError as e:
            print(e)

    @property
    def y(self):
        """:math:`y`-coordinate of `Vector2D`.

        Returns
        -------
        float
            :math:`y`-coordinate of `Vector2D`.

        """
        return self._v[1]

    @y.setter
    def y(self, value=float):
        """Set :math:`y`-coordinate of `Vector2D`.

        Parameters
        ----------
        value : float
            :math:`y`-coordinate of `Vector2D`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(int, float))
            try:
                self._v[1] = value
            except ValueError:
                self._v[1] = Qty(value, self._units)
        except TypeError as e:
            print(e)

    @property
    def components(self):
        """:math:`x, y` components of `Vector2D`.

        Returns
        -------
        :py:class:`~numpy:numpy.ndarray`
            2-element :py:class:`~numpy:numpy.ndarray` of [:math:`x, y`]
            components of `Vector2D`.

        """
        return self._v

    @components.setter
    def components(self, value=np.ndarray):
        """Set :math:`x, y` components of `Vector2D`

        Parameters
        ----------
        value : :py:class:`~numpy:numpy.ndarray`
            2-element :py:class:`~numpy:numpy.ndarray` of [:math:`x, y`]
            components of `Vector2D`.

        """
        try:
            value = value.magnitude
        except AttributeError:
            pass

        try:
            check_type(value, allowed_types=(np.ndarray,))
            for i, vi in enumerate(value):
                try:
                    check_type(vi, allowed_types=(int, float))
                    try:
                        self._v[i] = vi
                    except ValueError:
                        self._v[i] = Qty(vi, self._units)
                except TypeError as e:
                    print(e)
        except TypeError as e:
            print(e)

    def fix_minus_zero_components(self, epsilon=1.0e-10):
        """Set `Vector2D` components that are small, negative numbers to zero.

        Set `Vector2D` components that are negative and have absolute value
        less than `epsilon` to zero.

        Parameters
        ----------
        epsilon : float, optional
            Smallest allowed absolute value of any :math:`x,y` component.

        """
        v = self._v.tolist()
        for i, vi in enumerate(v[:]):
            if vi < 0 and abs(vi) < epsilon:
                try:
                    self._v[i] = 0.0
                except ValueError:
                    self._v[i] = Qty(0.0, self._units)

    def rezero_components(self, epsilon=1.0e-10):
        """Re-zero `Vector2D` components near zero.

        Set `Vector2D` components with absolute value less than `epsilon` to
        zero.

        Parameters
        ----------
        epsilon : float, optional
            Smallest allowed absolute value of any :math:`x,y` component.

        """
        v = self._v.tolist()
        for i, vi in enumerate(v[:]):
            if abs(vi) < epsilon:
                try:
                    self._v[i] = 0.0
                except ValueError:
                    self._v[i] = Qty(0.0, self._units)


class Quaternion(object):
    """Create a quaternion in :math:`R^3`

    Parameters
    ----------
    w, x, y, z : float, optional
        :math:`x, y, z` components of terminating point of vector in
        :math:`R^3` space relative to origin.

    Notes
    -----
    .. todo::

       add new methods for coordinate transformations


    """
    def __init__(self, w=None, x=None, y=None, z=None):

        raise RuntimeError('This class is in development and not '
                           'completely implemented or ready for use.')

        self._w = w
        self._x = x
        self._y = y
        self._z = z



# -*- coding: utf-8 -*-
"""
====================================================================
Abstract data structures for chemistry (:mod:`sknano.chemistry`)
====================================================================

.. currentmodule:: sknano.chemistry

Contents
========

.. autosummary::
   :toctree: generated/

   Atom
   Atoms

"""
from __future__ import division, print_function, absolute_import

__docformat__ = 'restructuredtext'

from ._atom import *
from ._atoms import *

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
"""
======================================================
Reference data package (:mod:`sknano.tools.refdata`)
======================================================

.. currentmodule:: sknano.tools.refdata

Contents
========

Periodic table of elements data
-------------------------------

.. autodata:: atomic_masses

.. autodata:: atomic_numbers

.. autodata:: element_symbols

.. autodata:: element_names

.. autodata:: CCbond

.. autodata:: C_C

.. autodata:: CHbond

.. autodata:: C_H

"""
from __future__ import division, absolute_import, print_function

__docformat__ = 'restructuredtext'

from ._constants import *
from ._periodic_table import *

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
"""Python toolkit for generating and analyzing nanostructure data"""

from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext'

import os
import sys
import subprocess

from setuptools import find_packages

from numpy.distutils.core import setup

if sys.version_info[:2] < (2, 7):
    raise RuntimeError("Python version 2.7 required.")

DISTNAME = 'scikit-nano'
DESCRIPTION = __doc__
LONG_DESCRIPTION = open('README.rst').read()
AUTHOR = 'Andrew Merrill'
AUTHOR_EMAIL = 'androomerrill@gmail.com'
MAINTAINER = AUTHOR
MAINTAINER_EMAIL = AUTHOR_EMAIL
URL = 'http://projects.geekcode.io/scikit-nano/doc'
DOWNLOAD_URL = 'https://github.com/androomerrill/scikit-nano'
KEYWORDS = 'nano nano-structures nanostructures nanotubes graphene LAMMPS XYZ'
LICENSE = 'BSD 2-Clause'
CLASSIFIERS = """\
Development Status :: 4 - Beta
Intended Audience :: Science/Research
License :: OSI Approved :: BSD License
Operating System :: MacOS :: MacOS X
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Programming Language :: Python
Programming Language :: Python :: 2.7
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Chemistry
Topic :: Scientific/Engineering :: Physics
Topic :: Scientific/Engineering :: Visualization
Topic :: Software Development
Topic :: Software Development :: Libraries :: Python Modules

"""

MAJOR = 0
MINOR = 2
MICRO = 24
ISRELEASED = True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)


def git_version():
    """Return the GIT version as a string."""
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = \
            subprocess.Popen(cmd,
                             stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


def get_version_info():
    # Adding the git rev number needs to be done inside
    # write_version_py(), otherwise the import of sknano.version messes
    # up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists('.git'):
        GIT_REVISION = git_version()
    elif os.path.exists('sknano/version.py'):
        # must be a source distribution, use existing version file
        # load it as a separate module to not load sknano/__init__.py
        import imp
        version = imp.load_source('sknano.version', 'sknano/version.py')
        GIT_REVISION = version.git_revision
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += '.dev-' + GIT_REVISION[:7]

    return FULLVERSION, GIT_REVISION


def write_version_py(filename='sknano/version.py'):
    cnt = """
# THIS FILE IS GENERATED FROM SCIKIT-NANO SETUP.PY
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
release = %(isrelease)s

if not release:
    version = full_version
"""
    FULLVERSION, GIT_REVISION = get_version_info()

    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION,
                       'git_revision': GIT_REVISION,
                       'isrelease': str(ISRELEASED)})
    finally:
        a.close()


def configuration(parent_package='', top_path=None):
    if os.path.exists('MANIFEST'):
        os.remove('MANIFEST')

    from numpy.distutils.misc_util import Configuration
    config = Configuration(None, parent_package, top_path)

    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)

    config.add_subpackage('sknano')

    return config


def setup_package():

    # Rewrite the version file everytime
    write_version_py()

    FULLVERSION, GIT_REVISION = get_version_info()

    setup_options = dict(
        name=DISTNAME,
        version=FULLVERSION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        url=URL,
        download_url=DOWNLOAD_URL,
        license=LICENSE,
        keywords=KEYWORDS,
        classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
        platforms=["Linux", "OS-X", "Unix", "Windows"],
        test_suite='nose.collector',
        configuration=configuration,
        packages=find_packages(exclude=['doc']),
        include_package_data=True,
        exclude_package_data={'': ['*.gif', '*.html', '*.ui']},
        zip_safe=False,
        install_requires=['numpy>=1.8', 'pint>=0.4'],
        entry_points={
            'console_scripts': [
                'nanogen = sknano.scripts.nanogen:main',
                'nanogenui = sknano.scripts.nanogenui:main',
            ]
        }
    )

    setup(**setup_options)

if __name__ == '__main__':
    setup_package()



# -*- coding: utf-8 -*-
"""
===================================================
Data analysis sub-package (:mod:`sknano.analysis`)
===================================================

.. currentmodule:: sknano.analysis

Contents
========

.. autosummary::
   :toctree: generated/

"""
from __future__ import division, print_function, absolute_import

__docformat__ = 'restructuredtext'

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
"""
=============================================================
Fullerene structure tools (:mod:`sknano.nanogen._fullerenes`)
=============================================================

.. currentmodule:: sknano.nanogen._fullerenes

"""
from __future__ import division, print_function, absolute_import
__docformat__ = 'restructuredtext'

#import itertools

#from fractions import gcd
#from collections import OrderedDict

#import numpy as np

try:
    from pint import UnitRegistry
    ureg = UnitRegistry()
    Qty = ureg.Quantity
except ImportError:
    Qty = None

#from ..chemistry import Atom

#param_units = {}

__all__ = ['Fullerene']


class Fullerene(object):
    u"""Class for creating interactive Fullerene objects.

    .. note::

       This class is under development and not functional! Attempting
       to use it raises a `RuntimeError`.

    The `fullerene isomer data
    <http://www.nanotube.msu.edu/fullerene/fullerene-isomers.html>`_
    were downloaded from the
    `The Nanotube Site <http://www.nanotube.msu.edu>`_
    at `Michigan State University <http://www.msu.edu>`_.

    Parameters
    ----------
    verbose : bool, optional
        verbose output

    Examples
    --------

    """
    def __init__(self, N=None, with_units=False, verbose=False):
        raise RuntimeError('This class is not yet implemented.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
#
# scikit-nano documentation build configuration file, created by
# sphinx-quickstart on Wed Oct 30 00:26:16 2013.
#
# This file is execfile()d with the current directory set to its containing dir

#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import re
import sys

# Check Sphinx version
import sphinx
if sphinx.__version__ < "1.1":
    raise RuntimeError("Sphinx 1.1 or newer required")

needs_sphinx = '1.1'

# -- General configuration ----------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../sphinxext'))
sys.path.insert(0, os.path.abspath('../sphinxext/numpydoc'))

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.mathjax', 'numpydoc',
              'sphinx.ext.intersphinx', 'sphinx.ext.coverage',
              'sphinx.ext.autosummary', 'sphinx.ext.inheritance_diagram',
              'sphinx.ext.todo', 'sphinx.ext.viewcode',
              'ipython_console_highlighting', 'sphinxarg.ext']

# Show todo items
todo_include_todos = True

# Determine if the matplotlib has a recent enough version of the
# plot_directive.
try:
    from matplotlib.sphinxext import plot_directive
except ImportError:
    use_matplotlib_plot_directive = False
else:
    try:
        use_matplotlib_plot_directive = (plot_directive.__version__ >= 2)
    except AttributeError:
        use_matplotlib_plot_directive = False

if use_matplotlib_plot_directive:
    extensions.append('matplotlib.sphinxext.plot_directive')
else:
    raise RuntimeError("You need a recent enough version of matplotlib")

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General substitutions.
project = u'scikit-nano'
copyright = u'2014, Andrew Merrill'

# The default replacements for |version| and |release|, also used in various
# other places throughout the built documents.
import sknano
version = re.sub(r'\.dev-.*$', r'.dev', sknano.__version__)
release = sknano.__version__

print("scikit-nano (VERSION {})".format(version))

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# List of documents that shouldn't be included in the build.
#unused_docs = []

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = "autolink"

# List of directories, relative to source directories, that shouldn't be
# searched for source files.
exclude_dirs = []

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
#pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#html_theme = 'sphinxdoc'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".

#themedir = '_theme'
html_theme = 'sknano'
html_theme_path = ['.']

html_theme_options = {
    "edit_link": False,
    "sidebar": "left",
    "sknano_logo": False,
    "rootlinks":
    [("http://github.com/androomerrill/scikit-nano", "scikit-nano"),
     ("http://projects.geekcode.io/scikit-nano/doc", "Docs")]}
html_logo = '_static/sknano_logo.png'
html_favicon = '_static/favicon.ico'
html_sidebars = {'index': 'indexsidebar.html'}

html_title = "%s v%s Reference Guide" % (project, version)
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

html_use_modindex = True
html_copy_source = False

#pngmath_use_preview = True
#pngmath_dvipng_args = ['-gamma', '1.5', '-D', '96', '-bg', 'Transparent']

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
html_file_suffix = '.html'

# Output file base name for HTML help builder.
htmlhelp_basename = 'scikit-nano-doc'

# -- Options for LaTeX output -------------------------------------------------

# The paper size ('letter' or 'a4').
latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, document class
# [howto/manual]).
latex_documents = [('index', 'scikit-nano-ref.tex',
                    'scikit-nano Reference Guide',
                    'Andrew Merrill', 'manual')]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# Additional stuff for the LaTeX preamble.
latex_preamble = r'''
\usepackage{amsmath}

\DeclareUnicodeCharacter{00A0}{\nobreakspace}

% In the parameters etc. sections, align uniformly, and adjust label emphasis
\usepackage{expdlist}
\let\latexdescription=\description
\let\endlatexdescription=\enddescription
\renewenvironment{description}%
{\begin{latexdescription}%
[\setleftmargin{60pt}\breaklabel\setlabelstyle{\bfseries\itshape}]}%
{\end{latexdescription}}

% Make Examples/etc section headers smaller and more compact
\makeatletter
\titleformat{\paragraph}{\normalsize\normalfont\bfseries\itshape}%
            {\py@NormalColor}{0em}{\py@NormalColor}{\py@NormalColor}
\titlespacing*{\paragraph}{0pt}{1ex}{0pt}
\makeatother

% Save vertical space in parameter lists and elsewhere
\makeatletter
\renewenvironment{quote}%
               {\list{}{\topsep=0pt%
                        \parsep \z@ \@plus\p@}%
                \item\relax}%
               {\endlist}
\makeatother

% Fix footer/header
\renewcommand{\chaptermark}[1]{\markboth{\MakeUppercase{\thechapter.\ #1}}{}}
\renewcommand{\sectionmark}[1]{\markright{\MakeUppercase{\thesection.\ #1}}}
'''

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
latex_use_modindex = False


# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'scikit-nano', u'scikit-nano Documentation',
     [u'Andrew Merrill'], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -----------------------------------------------


# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [('index', 'scikit-nano',
                      u'scikit-nano Documentation',
                      u'Andrew Merrill', 'scikit-nano',
                      'python toolkit',
                      'Miscellaneous')]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False


# -----------------------------------------------------------------------------
# Intersphinx configuration
# -----------------------------------------------------------------------------
intersphinx_mapping = {
    'python': ('http://docs.python.org/2', None),
    'numpy': ('http://docs.scipy.org/doc/numpy', None),
    'scipy': ('http://docs.scipy.org/doc/scipy/reference', None),
    'pint': ('http://pint.readthedocs.org/en/latest', None),
}

# -----------------------------------------------------------------------------
# Numpy extensions
# -----------------------------------------------------------------------------

# If we want to do a phantom import from an XML file for all autodocs
phantom_import_file = 'dump.xml'

# Generate plots for example sections
numpydoc_use_plots = True

# -----------------------------------------------------------------------------
# Autosummary
# -----------------------------------------------------------------------------

if sphinx.__version__ >= "0.7":
    import glob
    autosummary_generate = glob.glob("*.rst")

# -----------------------------------------------------------------------------
# Coverage checker
# -----------------------------------------------------------------------------
coverage_ignore_modules = r"""
    """.split()
coverage_ignore_functions = r"""
    test($|_) (some|all)true bitwise_not cumproduct pkgload
    generic\.
    """.split()
coverage_ignore_classes = r"""
    """.split()

coverage_c_path = []
coverage_c_regexes = {}
coverage_ignore_c_items = {}


#------------------------------------------------------------------------------
# Plot
#------------------------------------------------------------------------------
plot_pre_code = """
import numpy as np
import scipy as sp
np.random.seed(123)
"""
plot_include_source = True
plot_formats = [('png', 96), 'pdf']
plot_html_show_formats = False

import math
phi = (math.sqrt(5) + 1)/2

font_size = 13*72/96.0  # 13 px

plot_rcparams = {
    'font.size': font_size,
    'axes.titlesize': font_size,
    'axes.labelsize': font_size,
    'xtick.labelsize': font_size,
    'ytick.labelsize': font_size,
    'legend.fontsize': font_size,
    'figure.figsize': (3*phi, 3),
    'figure.subplot.bottom': 0.2,
    'figure.subplot.left': 0.2,
    'figure.subplot.right': 0.9,
    'figure.subplot.top': 0.85,
    'figure.subplot.wspace': 0.4,
    'text.usetex': False,
}

if not use_matplotlib_plot_directive:
    import matplotlib
    matplotlib.rcParams.update(plot_rcparams)

# -----------------------------------------------------------------------------
# Source code links
# -----------------------------------------------------------------------------

import inspect
from os.path import relpath, dirname

for name in ['sphinx.ext.linkcode', 'linkcode', 'numpydoc.linkcode']:
    try:
        __import__(name)
        extensions.append(name)
        break
    except ImportError:
        pass
else:
    print("NOTE: linkcode extension not found -- no links to source generated")


def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object
    """
    if domain != 'py':
        return None

    modname = info['module']
    fullname = info['fullname']

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split('.'):
        try:
            obj = getattr(obj, part)
        except:
            return None

    try:
        fn = inspect.getsourcefile(obj)
    except:
        fn = None
    if not fn:
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except:
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.findsource(obj)
    except:
        lineno = None

    if lineno:
        linespec = "#L%d" % (lineno + 1)
    else:
        linespec = ""

    fn = relpath(fn, start=dirname(sknano.__file__))

    if 'dev' in sknano.__version__:
        return("https://github.com/androomerrill/scikit-nano/blob/" +
               "dev/sknano/{}{}".format(fn, linespec))
    else:
        return("https://github.com/androomerrill/scikit-nano/blob/" +
               "v{}/sknano/{}{}".format(sknano.__version__, fn, linespec))


class Mock(object):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    @classmethod
    def __getattr(cls, name):
        if name in ('__file__', '__path__'):
            return '/dev/null'
        elif name[0] == name[0].upper():
            mockType = type(name, (), {})
            mockType.__module__ = __name__
            return mockType
        else:
            return Mock()

MOCK_MODULES = []
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()



# -*- coding: utf-8 -*-
"""
=============================================
Command-line scripts (:mod:`sknano.scripts`)
=============================================

.. currentmodule:: sknano.scripts

Contents
========

.. autosummary::
   :toctree: generated/

   nanogen
   nanogenui

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from .nanogen import *
from .nanogenui import NanoGen

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
"""
===============================================================
Package for generating nano-structures (:mod:`sknano.nanogen`)
===============================================================

.. currentmodule:: sknano.nanogen

Collection of modules for generating nano-structures including graphene,
bi-layer graphene, :math:`N`-layer graphene,
single-walled nanotubes (:abbr:`SWNTs (single-walled nanotubes)`),
multi-walled nanotubes (:abbr:`MWNTs (multi-walled nanotubes)`),
and bundles of :abbr:`SWNTs` and :abbr:`MWNTs`.

See the individual class doc pages for detailed documentation and
examples on how to use them.

.. note::
   The default basis atoms are both carbon and therefore the default
   nano-structures are carbon nano-structures. However,
   you can use any atom in the periodic table of elements as a basis atom.

.. versionadded:: 0.2.23
   `UnrolledNanotubeGenerator` implemented.

.. versionadded:: 0.2.20
   `MWNTBundleGenerator` implemented.

.. versionchanged:: 0.2.20
   `MWNTGenerator` no longer generates MWNT *bundles*, only *single* MWNTs.
   To generate bundled MWNT structure data, use the `MWNTBundleGenerator`
   class.

.. versionadded:: 0.2.8
   `MWNTGenerator` implemented.

.. versionadded:: 0.2.6
   `GrapheneVacancyGenerator` and `NanotubeVacancyGenerator` implemented.

.. versionadded:: 0.2.4
   `NanotubeBundleGenerator` implemented.

.. seealso::

   CLI script: :mod:`sknano.scripts.nanogen`
      The :mod:`~sknano.scripts.nanogen` script provides a
      :abbr:`CLI (command-line interface)` to the `nanogen` structure
      generator classes.

   GUI front-end: :mod:`sknano.scripts.nanogenui`
      The :mod:`~sknano.scripts.nanogenui` script launches a
      :abbr:`GUI (graphical user interface)` front-end to the
      `nanogen` structure generator classes.

Contents
========

Classes for creating abstract physical representations of nano-structures
--------------------------------------------------------------------------
.. autosummary::
   :toctree: generated/

   Fullerene
   Graphene
   GraphenePrimitiveCell
   Nanotube
   NanotubeBundle

Classes for generating fullerene structure data
------------------------------------------------
.. autosummary::
   :toctree: generated/

   FullereneGenerator
   FullereneBravaisLatticeGenerator

Classes for generating graphene structure data
-----------------------------------------------
.. autosummary::
   :toctree: generated/

   GrapheneGenerator
   BiLayerGrapheneGenerator
   UnrolledNanotubeGenerator

Classes for generating nanotube structure data
-----------------------------------------------
.. autosummary::
   :toctree: generated/

   NanotubeGenerator
   NanotubeBundleGenerator
   UnrolledNanotubeGenerator
   MWNTGenerator
   MWNTBundleGenerator
   TubeGen

Classes for generating nano-structure data with vacancies
----------------------------------------------------------
.. autosummary::
   :toctree: generated/

   GrapheneVacancyGenerator
   NanotubeVacancyGenerator

Base classes for creating new structure data generators
--------------------------------------------------------
.. autosummary::
   :toctree: generated/

   StructureGenerator
   VacancyGenerator

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from ._defect_generators import *
from ._fullerene_bravais_lattice_generators import *
from ._fullerene_generators import *
from ._fullerenes import *
from ._graphene_defect_generators import *
from ._graphene_generators import *
from ._graphene import *
from ._nanotube_defect_generators import *
from ._nanotube_bundle_generators import *
from ._nanotube_generators import *
from ._nanotube_junction_generators import *
from ._nanotubes import *
from ._structure_generator import *
from ._tubegen import *
from ._twisted_bundle_generators import *

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
#!/usr/bin/env python
"""
===================================================
Command line script (:mod:`sknano.scripts.nanogen`)
===================================================

CLI to :mod:`sknano.nanogen` tools.

.. currentmodule:: sknano.scripts.nanogen

This module allows for easy structure generation from the command line.

The :mod:`sknano.nanogen` package provides a python class
(:py:class:`~sknano.nanogen.TubeGen`) which wraps the
functionality of the TubeGen ([TG]_) binary executable.
TubeGen generates structure data for nanotubes and graphene.

You must have a compiled binary executable of TubeGen available in the
system :envvar:`PATH` and callable as :program:`tubegen`.

.. seealso::

   :py:class:`~sknano.nanogen.TubeGen`

   :py:class:`~sknano.nanogen.GrapheneGenerator`

   :py:class:`~sknano.nanogen.NanotubeBundleGenerator`

   :py:class:`~sknano.nanogen.MWNTGenerator`

.. code-block:: python

   > nanogen --help
   usage: nanogen [-h]
                  [--format {pdb,pdb-pbc,cif,bgf,pov,wien,xyz,gaussian,
                  gaussian-pbc}]
                  [--units {angstrom,bohr}] [--element1 ELEMENT1]
                  [--element2 ELEMENT2] [--bond BOND] [--gutter X Y Z]
                  [--shape {hexagonal,cubic,planar}] [--chirality n m]
                  [--relax-tube {yes,no}]
                  [--cell-count X Y Z | --tube-length L_tube] [--xyz2data]
                  [--add-atomtype ELEMENT ATOM-TYPE]
                  [--generator {tubegen,nanogen}]

   optional arguments:
     -h, --help            show this help message and exit
     --format {pdb,pdb-pbc,cif,bgf,pov,wien,xyz,gaussian,gaussian-pbc}
     --units {angstrom,bohr}
     --element1 ELEMENT1
     --element2 ELEMENT2
     --bond BOND
     --gutter X Y Z
     --shape {hexagonal,cubic,planar}
     --chirality n m
     --relax-tube {yes,no}
     --cell-count X Y Z
     --tube-length L_tube
     --xyz2data
     --add-atomtype ELEMENT ATOM-TYPE
     --generator {tubegen,nanogen}

.. autofunction:: nanogen

Examples
--------

::

    > nanogen --chirality 10 10 --cell-count 1 1 5

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import argparse
import sys

from ..chemistry import Atom
from ..nanogen import GrapheneGenerator, NanotubeBundleGenerator, \
    TubeGen, format_ext
from ..structure_io import XYZ2DATAConverter
from ..tools.refdata import CCbond

__all__ = ['nanogen']


def argparser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--format', default='xyz', dest='fmt',
                        choices=tuple([fmt for fmt in format_ext.iterkeys()]),
                        help='output file format (default: %(default)s)')
    parser.add_argument('--units', default='angstrom',
                        choices=('angstrom', 'bohr'),
                        help='units (default: %(default)s)')
    parser.add_argument('--element1', default='C',
                        help='element symbol or atomic number of basis '
                        'atom 1 (default: %(default)s)')
    parser.add_argument('--element2', default='C',
                        help='element symbol or atomic number of basis '
                        'atom 2 (default: %(default)s)')
    parser.add_argument('--bond', type=float, default=CCbond,
                        help='element1-element2 bond length in Angstroms '
                        '(default: %(default)s)')
    parser.add_argument('--gutter', nargs=3, type=float,
                        metavar=('X', 'Y', 'Z'), default=(1.6735, 1.6735, 0),
                        help='cell gutter (default: %(default)s)')
    #parser.add_argument('--structure', default='nanotube',
    #        choices=('nanotube', 'graphene', 'hexagonal-bundle'),
    #        help='crystal structure (default: %(default)s)')
    parser.add_argument('--shape', default='hexagonal',
                        choices=('hexagonal', 'cubic', 'planar'),
                        help='crystal structure (default: %(default)s)')
    parser.add_argument('--chirality', nargs=2, type=int, metavar=('n', 'm'),
                        default=(10, 10), help='chiral indices (n,m) '
                        '(default: %(default)s)')
    parser.add_argument('--relax-tube', default='yes', choices=('yes', 'no'),
                        help='select tube relaxation (default: %(default)s)')

    parser_group = parser.add_mutually_exclusive_group()
    parser_group.add_argument('--cell-count', nargs=3, type=int,
                              metavar=('X', 'Y', 'Z'), default=(1, 1, 1),
                              help='number of unit cells '
                              '(default: %(default)s)')
    parser_group.add_argument('--tube-length', type=float,
                              metavar='L_tube', default=None,
                              help='tube length in nanometers '
                              '(default: %(default)s)')

    parser.add_argument('--xyz2data', action='store_true',
                        help='convert xyz to LAMMPS data format')
    parser.add_argument('--add-atomtype', nargs=2, action='append',
                        metavar=('ELEMENT', 'ATOM-TYPE'),
                        dest='new_atomtypes',
                        help='add new atom to structure data.')

    parser.add_argument('--generator', choices=('tubegen', 'nanogen'),
                        default='tubegen', help='nano-structure generator '
                        '(default: %(default)s)')

    return parser


def nanogen(fmt='xyz', units='angstrom', element1='C', element2='C',
            bond=1.421, gutter=(1.6735, 1.6735, 0), shape='hexagonal',
            chirality=(10, 10), relax_tube='yes', cell_count=(1,1,1),
            tube_length=None, xyz2data=False, new_atomtypes=None,
            generator='tubegen'):
    """Generate nano-structure data.

    Parameters
    ----------
    fmt : str, optional
        output file format
    units : str, optional
    element1 : str, optional
        element symbol or atomic number of basis atom 1
    element2 : str, optional
        element symbol or atomic number of basis atom 2
    bond : float, optional
        element1-element2 bond length in ``units``
    gutter : sequence, optional
        cell gutter
    shape : {'hexagonal', 'cubic', 'planar'}, optional
        crystal structure
    chirality : sequence, optional
        chiral indicies (n, m)
    relax_tube : {'yes', 'no'}, optional
    cell_count : sequence, optional
        number of unit cells
    tube_length : {None, float}, optional
        tube length in **nanometers**
    xyz2data : bool, optional
        convert xyz to LAMMPS data format
    new_atomtypes : {None, sequence}, optional
        add new (element, atomtype) to LAMMPS data
    generator : {'tubegen', 'nanogen'}, optional
        nano-structure generator

    """
    if generator == 'tubegen':
        if shape == 'planar':
            gutter = (0.0, gutter[1], gutter[2])

        tubegen = TubeGen(fmt=fmt, units=units, element1=element1,
                          element2=element2, bond=bond, gutter=gutter,
                          shape=shape, chirality=chirality,
                          cell_count=cell_count, tube_length=tube_length,
                          relax_tube=relax_tube)

        tubegen.generate()
        tubegen.cleanup()

        if tube_length is not None:
            tubegen.fix_length()

        if xyz2data:
            if fmt == 'xyz':
                xyzconverter = XYZ2DATAConverter(tubegen.output, pad_box=True)
                if new_atomtypes is not None:
                    for element, atomtype in new_atomtypes:
                        atom = Atom(element, atomtype=int(atomtype))
                        xyzconverter.add_atomtype(atom=atom)
                xyzconverter.convert()
            else:
                print("can't convert {} format to data. ".format(fmt) +
                      "Must be xyz format.")
    else:
        n, m = chirality
        nx, ny, nz = cell_count

        if shape == 'planar':
            generator = GrapheneGenerator()
        else:
            generator = NanotubeBundleGenerator(n, m, nx=nx, ny=ny, nz=nz,
                                                Lz=tube_length,
                                                bundle_geometry=shape)

        if fmt not in ('xyz', 'data'):
            fmt = 'xyz'

        generator.save_data(structure_format=fmt)


def main():

    args = argparser().parse_args()
    nanogen(**vars(args))

if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
"""
=============================================================================
Base classes for structure data (:mod:`sknano.structure_io._structure_data`)
=============================================================================

.. currentmodule:: sknano.structure_io._structure_data

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from abc import ABCMeta, abstractmethod
from collections import OrderedDict

from ..chemistry import Atoms
from sknano.version import short_version as version

default_comment_line = \
    'Structure data generated using scikit-nano version {}'.format(version)
default_structure_format = 'xyz'
supported_structure_formats = ('xyz', 'data')

__all__ = ['StructureData',
           'StructureReader',
           'StructureWriter',
           'StructureConverter',
           'StructureFormat',
           'StructureDataError',
           'StructureReaderError',
           'StructureWriterError',
           'StructureConverterError',
           'StructureFormatError',
           'default_comment_line',
           'default_structure_format',
           'supported_structure_formats']


class StructureData(object):
    """Base class defining common properties for structure data.

    Parameters
    ----------
    fpath : {None, str}, optional

    """
    def __init__(self, fpath=None):
        self._comment_line = default_comment_line
        self._fpath = fpath
        self._structure_atoms = Atoms()
        self._properties = OrderedDict()

    @property
    def structure_atoms(self):
        """:class:`~sknano.chemistry.Atoms` instance."""
        return self._structure_atoms

    @property
    def atoms(self):
        """Alias for :attr:`~StructureData.structure_atoms`."""
        return self.structure_atoms

    @property
    def comment_line(self):
        """Comment line."""
        return self._comment_line

    @comment_line.setter
    def comment_line(self, value=str):
        """Set the comment line string.

        Parameters
        ----------
        value : str

        """
        self._comment_line = value

    @property
    def fpath(self):
        """File path"""
        return self._fpath

    @fpath.setter
    def fpath(self, value=str):
        """Set the file path string.

        Parameters
        ----------
        value : str

        """
        self._fpath = value

    @fpath.deleter
    def fpath(self):
        """Delete file path string"""
        del self._fpath

    @property
    def Natoms(self):
        """Number of atoms in :class:`~sknano.chemistry.Atoms`"""
        return self._structure_atoms.Natoms

    @property
    def properties(self):
        """:class:`~python:collections.OrderedDict` of structure properties."""
        return self._properties


class StructureReader(StructureData):
    """Abstract base class for reading structure data.

    Parameters
    ----------
    fpath : str
        structure file

    """
    __metaclass__ = ABCMeta

    def __init__(self, fpath=None):
        super(StructureReader, self).__init__(fpath=fpath)

    @abstractmethod
    def read(self):
        """Read structure data from file"""
        return NotImplementedError('Subclasses of `StructureReader` need to '
                                   'implement the `read` method.')


class StructureWriter(object):
    """Abstract base class for writing structure data."""
    __metaclass__ = ABCMeta

    @abstractmethod
    def write(self):
        """Write structure data to file"""
        return NotImplementedError('Subclasses of `StructureWriter` need to '
                                   'implement the `write` method.')


class StructureConverter(object):
    """Abstract base class for converting structure data.

    Parameters
    ----------
    infile : str
    outfile : str

    """
    __metaclass__ = ABCMeta

    def __init__(self, infile=None, outfile=None):
        self._infile = infile
        self._outfile = outfile

    @property
    def infile(self):
        """Return `infile`"""
        return self._infile

    @property
    def outfile(self):
        """Return `outfile`"""
        return self._outfile

    @abstractmethod
    def convert(self):
        """Convert structure data from one format to another format."""
        return NotImplementedError('Subclasses of `StructureConverter` need '
                                   'to implement the `convert` method.')


class StructureFormat(object):
    """Base class for containing structure format properties"""
    def __init__(self):
        self._properties = OrderedDict()

    @property
    def properties(self):
        """:class:`~python:collections.OrderedDict` of format properties."""
        return self._properties


class StructureFormatter(object):
    """Abstract base class for formatting `Atom` attributes."""

    def format(self):
        """Convert `Atom` attributes based on `StructureFormat`"""
        pass


class StructureDataError(Exception):
    """Base class for `StructureData` exceptions."""
    pass


class StructureReaderError(StructureDataError):
    """Exception raised for `StructureReader` errors.

    Parameters
    ----------
    msg : str
        Error message.

    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class StructureWriterError(StructureDataError):
    """Exception raised for `StructureWriter` errors.

    Parameters
    ----------
    msg : str
        Error message.

    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class StructureConverterError(StructureDataError):
    """Exception raised for `StructureConverter` errors.

    Parameters
    ----------
    msg : str
        Error message.

    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class StructureFormatError(StructureDataError):
    """Exception raised for `StructureFormat` errors.

    Parameters
    ----------
    msg : str
        Error message.

    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)



# -*- coding: utf-8 -*-
"""
==========================================================
Tools for structure data I/O (:mod:`sknano.structure_io`)
==========================================================

.. currentmodule:: sknano.structure_io

Contents
========

Classes for handling `LAMMPS data` structure data format
--------------------------------------------------------
.. autosummary::
   :toctree: generated/

   DATAData
   DATAReader
   DATAWriter
   DATA2XYZConverter
   DATAFormat
   DATAError

Classes for handling `xyz` structure data format
-------------------------------------------------
.. autosummary::
   :toctree: generated/

   XYZData
   XYZReader
   XYZWriter
   XYZ2DATAConverter
   XYZFormat

Base classes for creating new structure data readers/writers/converters
------------------------------------------------------------------------
.. autosummary::
   :toctree: generated/

   StructureData
   StructureReader
   StructureWriter
   StructureConverter
   StructureFormat

Base custom exception classes for handling errors
--------------------------------------------------
.. autosummary::
   :toctree: generated/

   StructureDataError
   StructureReaderError
   StructureWriterError
   StructureConverterError
   StructureFormatError

Sub-packages
=============
.. autosummary::
   :toctree: generated/

   atoms

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from ._lammps_data_format import *
from ._xyz_format import *
from ._structure_data import *

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
"""
=========================================================
Reference data (:mod:`sknano.tools.refdata._constants`)
=========================================================

.. currentmodule:: sknano.tools.refdata._constants

"""
from __future__ import division, print_function, absolute_import

C_C = CCbond = 1.421  # angstroms
C_H = CHbond = 1.09  # angstroms
dVDW = 3.4  # angstroms

__all__ = ['C_C', 'CCbond', 'C_H', 'CHbond', 'dVDW']
